; programme to use pic 16f88 as i2c slave driving lcd
;date 7/2/2009 - working 9/2/2009 - bugs out 13/2/2009

	#include p16f88.inc
	list p=p16f88
;
uSec10 equ h'20'		;delay variables
ms1 equ h'21'
sec1 equ h'22'
ms10 equ h'23'


delay_byte equ h'24'
databyte equ h'25'		;register for data to o/p or i/p
buff_size equ 8		;number of internal addresses
offset equ h'26'		;internal address pointer
addr0 equ h'27'		;internal memory locations
addr1 equ h'28'		;more or less as required or ram permits
addr2 equ h'29'
addr3 equ h'2a'
addr4 equ h'2b'
addr5 equ h'2c'
addr6 equ h'2d'
addr7 equ h'2e'

i2cdata equ h'2f'		;display variables
tempw equ h'30'
status_temp equ h'31'
fsr_temp equ h'32'
address equ b'01110000'	;base address of slave

rs equ 4		;porta - disp register select
en equ 7		;porta - display enable
	
	__config config1, _lvp_off & _boden_off & _Mclr_off & _wdt_off & _intrc_io ;& _cp_all

	org 0
	goto start
	org 4
	goto interrupt
	org 6
start:
	clrf porta		;get rid of any leftover data
	clrf portb
	bcf status,irp
	bcf status,rp1
	bsf status,rp0		;bank 1
	clrf ansel
	clrf trisa		;port a outputs
	movlw h'fe'
	movwf trisb		;port b inputs 7..1, portb 0 output
	bsf trisb,1		;scl and sda as inputs for i2c
	bsf trisb,4		;to be doubly sure
	movlw b'01100000'	;set int rc osc to 4MHz
	movwf osccon
	clrf sspstat		;clear ssp status register
	bcf status,rp0		;select bank 0
	movf portb,w
	andlw b'00001100'	;mask for address bits
	movwf tempw		;save bits in temp register
	bcf status,c		;clear carry bit
	rrf tempw,w		;shit bits to pos 1 and 2
	iorlw address		;combine with base address
	bsf status,rp0		;bank 1

	

	movwf sspadd
	bsf pie1,sspie		;enable interrupts
	bcf status,rp0		;bank 0
	bcf pir1,sspif
	movlw b'00110110'	;set up for slave read no interrupt
	movwf sspcon
	bsf intcon,peie		
	bsf sspcon,sspen
	bsf intcon,gie

;===================================================================
;                    end of pic setup routine
;  		beginning of main-loop initialisation
;===================================================================
	clrf offset		;set up indirect address for slave regs


;===================================================================
;===================================================================
;       main slave function programme starts here
;===================================================================
;===================================================================

	movlw 5
	call delayms		;delay 1 second
	call init		;set up display
	clrf addr0	
	clrf addr1
	clrf i2cdata
main_loop:			;not used here
	
wait_data:
	
	movlw 0			;for some reason movf addr0,f
	subwf addr1,w		;doesn't work all the time
	btfsc status,z
	goto wait_data		;no data yet
	movf addr0,w		;get data byte in w reg
	btfsc addr1,0		;check if char or instr, 1 = char
	goto out_char		;send the char to the display
	btfsc addr1,2		;check if a system reset b2 = 1
	goto dispres
	call wrins		;output an instruction
	clrf addr1		;tell master data is used
	goto wait_data
dispres:
	call init
	clrf addr1
	goto wait_data
out_char:
	call wrchr
	clrf addr1
	goto wait_data	

	goto main_loop			;not used here



;===================================================================
;	start of main_loop routines etc.
;===================================================================

;===================================================================
;	set up display as 2 lines x 16 chars 5x8 matrix
;===================================================================
init:
	clrf porta
	movlw 100		;power up pause
	call delayms
	movlw b'00000010'	;4 bit mode
	movwf porta
	call pulse_en
 	movlw 5
	call delay10
	movlw b'00101100'	;
	call wrins
	movlw 5
	call delay10
	movlw b'00001100'	;display control
	call wrins
	movlw 5
	call delay10
	movlw b'00000001'	;clear display
	call wrins
	movlw 2
	call delayms
	movlw b'00000110'	;entry mode
	call wrins
	movlw 60
	call delay10
	return


;===================================================================
;	output an instruction to the display
;===================================================================
wrins:
	movwf i2cdata		;save instruction
;	movlw h'28'		;check to see if single or double line
;	andwf i2cdata,w		;mask off unwanted bits
;	bcf portb,0		;pre clear contrast control pin rb0
;	sublw h'28'		;compare bytes
;	btfss status,z	
;	bsf portb,0		;single, so set contrast pin

	swapf i2cdata,w		;put high nibble in w low nibble
	andlw h'0f'		;mask off high nibble
	movwf porta
	bcf porta,rs
	call pulse_en		;activate en pulse
	movf i2cdata,w		;get low nibble of instruction
	andlw h'0f'		;mask off high nibble
	movwf porta
	bcf porta,rs
	call pulse_en
	movlw 4
	call delay10
	return

;===================================================================
;	output a character to the display
;===================================================================
wrchr:
	movwf i2cdata
	swapf i2cdata,w		;put high nibble in w low nibble
	andlw h'0f'		;mask off high nibble
	movwf porta
	bsf porta,rs		;set rs to char mode
	call pulse_en		;activate en pulse
	movf i2cdata,w		;get low nibble of char
	andlw h'0f'		;mask off high nibble
	movwf porta
	bsf porta,rs
	call pulse_en
	bcf porta,rs
	movlw 4
	call delay10
	return
;===================================================================
;	pulse the en line to set the inst/char
;===================================================================
pulse_en:

	nop
	bsf porta,en		;en high
	nop
	bcf porta,en		;en low
	nop

	return
;===================================================================
;delay routines - set for 4MHz clock speed
;===================================================================


delay10:			;10uSec per loop
	movwf usec10		;come in with uSx10 count in w reg
delay10a:
	nop			;pad out to 10uSecs
	nop
	nop
	nop
	nop
	nop
	decfsz usec10,f
	goto delay10a		;loop count 10uSec loops
	return

delayms:
	movwf ms1		;come in with mSec count in W reg
delaymsa:
	movlw 100		;set up for 100 x 10uSec = 1 mSec
	call delay10
	decfsz ms1,f
	goto delaymsa
	return

delayms10:			;come in with mSx10 count in W reg
	movwf ms10
delayms10a:
	movlw 10
	call delayms
	decfsz ms10,f
	goto delayms10a
	return

delaysec:
	movwf sec1		;come in with seconds count in W reg
delayseca:
	movlw 100		;1000 mSec
	call delayms10			
	decfsz sec1,f
	goto delayseca
	return

;========================================================================
;	End of Main Loop Routines etc.
;	Beginning of interrupt routines
;========================================================================
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;========================================================================
;	this is the data i/o section - i2c slave main routine
;========================================================================
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

interrupt:
	bcf status,rp0		;bank 0
	bcf pir1,sspif
	movwf tempw		;save contents of w reg
	movf status,w
	movwf status_temp	;save contents of status reg
	movf fsr,w
	movwf fsr_temp		;save fsr register
	movf sspbuf,w		;clear data buffer 
	movwf databyte		;save in databyte
	bsf status,rp0		;bank 1
	btfsc sspstat,d_a	;check if data or address
	goto data_mode

;========================================================================
;	get and process mode byte	
;========================================================================

	btfsc sspstat,r_w	;check if we need a mode byte
	goto send
	call inbyte		;read mode byte
	movlw buff_size		;get number of internal addresses
	subwf databyte,w	;check if valid address
	btfsc status,c		;valid if carry is 0 
	clrf databyte		;if bad address, assume address 0
	movf databyte,w		;good address, so put it in <offset>
	movwf offset
	goto int_end

data_mode:

	btfss sspstat,r_w	;check if read or write
	goto receive		;input data
send:
	bcf status,rp0		;bank 0
	movlw addr0		;physical location of address block	
	addwf offset,w		;point to required address
	movwf fsr		;put address in indirect address mode
	movf indf,w		;get data to be output
	btfsc sspcon,ckp	;check if another byte
	goto int_end		;no so finish
	movwf sspbuf		;send data to master
	nop			;delay needed to correct timing
	nop
	nop
	nop
	nop
	bsf sspcon,ckp
	bcf pir1,sspif
	incf offset,f
	movlw buff_size
	subwf offset,w
	btfsc status,z
	clrf offset
	goto int_end

receive:
	bcf status,rp0		;bank 0
	movlw addr0		;physical location of address block
	addwf offset,w		;point to required address
	movwf fsr		;put address in indirect address mode
	movf databyte,w
	movwf indf		;put data in appropriate address
	incf offset,f		;automatic increment of address
	movlw buff_size
	subwf offset,w		;check if exceeded address limit
	btfsc status,z
	clrf offset		;back to start of addresses
	goto int_end	

;========================================================================

int_end:


	bcf status,rp0		;reset bank 0
	movf fsr_temp,w
	movwf fsr		
	movf status_temp,w	;restore temps
	movwf status
	movf tempw,w
	bcf pir1,sspif		;clear the interrupt flag
	retfie  		

inbyte:
	bcf status,rp0		;bank 0
	btfss pir1,sspif	;wait for databyte in
	goto inbyte
	movf sspbuf,w		;read data buffer and put in databyte reg
	movwf databyte
	bcf pir1,sspif		;clear interrupt flag
	return			
here:
	movlw h'ff'		;debug routine 1
	xorwf porta,f
	goto here
debug:
	bsf porta,7		;debug routine 2
	bsf porta,4
	bcf porta,7
	bcf porta,4
	goto debug

	end			;mandatory instruction